import argparse
from pathlib import Path
from typing import Callable

import numpy as np
import torch
from torch.utils.data import DataLoader

from elasticai.creator.examples.assets.determine_fixed_point.pems_dataset import (
    PeMSDataset,
)
from elasticai.creator.resource_utils import PathType
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import (
    Linear1dTranslationArgs,
    LSTMTranslationArgs,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)


def read_commandline_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--build_dir", required=True, type=Path)
    return parser.parse_args()


class LSTMModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.lstm = torch.nn.LSTM(input_size=1, hidden_size=20, batch_first=True)
        self.linear = torch.nn.Linear(in_features=20, out_features=1)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        return self.linear(self.lstm(x)[0])

    def get_weights_and_bias(self) -> list[np.ndarray]:
        return [
            self.lstm.weight_ih_l0.detach().numpy().squeeze(),
            self.lstm.weight_hh_l0.detach().numpy(),
            self.lstm.bias_ih_l0.detach().numpy(),
            self.lstm.bias_hh_l0.detach().numpy(),
            self.linear.weight.data.detach().numpy().squeeze(),
            self.linear.bias.data.detach().numpy(),
        ]


def determine_optimal_fixed_point_factory(
    model: LSTMModel,
    saved_model_path: PathType,
    data_path: PathType,
    sensor_idx: int,
    batch_size: int,
    datapoints_per_sample: int,
) -> Callable[[float | int], FixedPoint]:
    def train_test_split(
        dataset_path: PathType, datapoints_per_sample: int, sensor_idx: int
    ) -> tuple[PeMSDataset, PeMSDataset]:
        n_three_week_datapoints = 12 * 24 * 7 * 3
        train = PeMSDataset(
            dataset_path,
            datapoints_per_sample,
            sensor_idx,
            raw_data_slice=(0, n_three_week_datapoints),
        )
        test = PeMSDataset(
            dataset_path,
            datapoints_per_sample,
            sensor_idx,
            raw_data_slice=(n_three_week_datapoints, None),
        )
        return train, test

    _, test_data = train_test_split(
        dataset_path=data_path,
        datapoints_per_sample=datapoints_per_sample,
        sensor_idx=sensor_idx,
    )
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True)

    model.load_state_dict(torch.load(saved_model_path))
    model.eval()
    weights_and_bias = model.get_weights_and_bias()

    total_bits = 16
    frac_bits_choices = [4, 6, 8, 10, 12, 14, 16]
    frac_bits_mse = []

    for frac_bits in frac_bits_choices:
        fixed_point_one = 2**frac_bits
        fp_args = dict(total_bits=total_bits, frac_bits=frac_bits)

        @np.vectorize
        def to_fixed_point(value: float) -> int:
            return FixedPoint.get_factory(**fp_args)(value).to_signed_int()

        @np.vectorize
        def to_float(value: int) -> float:
            return float(FixedPoint.from_int(value, **fp_args, signed_int=True))

        def fixed_point_matmul(a: np.ndarray, b: np.ndarray) -> np.ndarray:
            return (np.matmul(a, b) * 1.0 / fixed_point_one).astype(int)

        #        def to_fixed_point(x: np.ndarray) -> np.ndarray:
        #            return (x * fixed_point_one).astype(int)
        #
        #        def to_float(x: np.ndarray) -> np.ndarray:
        #            return x * 1.0 / fixed_point_one
        #
        #        def fixed_point_matmul(a: np.ndarray, b: np.ndarray) -> np.ndarray:
        #            return (np.matmul(a, b) * 1.0 / fixed_point_one).astype(int)

        def sigmoid(x: np.ndarray) -> np.ndarray:
            return to_fixed_point(1 / (1 + np.exp(-to_float(x))))

        def tanh(x: np.ndarray) -> np.ndarray:
            return to_fixed_point(np.tanh(to_float(x)))

        def qlstm_cell(
            x: np.ndarray,
            h_prev: np.ndarray,
            c_prev: np.ndarray,
            weights_and_bias: list[np.ndarray],
        ) -> tuple[np.ndarray, np.ndarray]:
            x = to_fixed_point(np.ones([1, 1]) * x)

            w_ii = to_fixed_point(np.expand_dims(weights_and_bias[0][:20], axis=1))
            w_if = to_fixed_point(np.expand_dims(weights_and_bias[0][20:40], axis=1))
            w_ig = to_fixed_point(np.expand_dims(weights_and_bias[0][40:60], axis=1))
            w_io = to_fixed_point(np.expand_dims(weights_and_bias[0][60:80], axis=1))

            w_hi = to_fixed_point(weights_and_bias[1][:20])
            w_hf = to_fixed_point(weights_and_bias[1][20:40])
            w_hg = to_fixed_point(weights_and_bias[1][40:60])
            w_ho = to_fixed_point(weights_and_bias[1][60:80])

            b_ii = to_fixed_point(np.expand_dims(weights_and_bias[2][:20], axis=1))
            b_if = to_fixed_point(np.expand_dims(weights_and_bias[2][20:40], axis=1))
            b_ig = to_fixed_point(np.expand_dims(weights_and_bias[2][40:60], axis=1))
            b_io = to_fixed_point(np.expand_dims(weights_and_bias[2][60:80], axis=1))

            b_hi = to_fixed_point(np.expand_dims(weights_and_bias[3][:20], axis=1))
            b_hf = to_fixed_point(np.expand_dims(weights_and_bias[3][20:40], axis=1))
            b_hg = to_fixed_point(np.expand_dims(weights_and_bias[3][40:60], axis=1))
            b_ho = to_fixed_point(np.expand_dims(weights_and_bias[3][60:80], axis=1))

            f_t = sigmoid(
                fixed_point_matmul(w_if, x)
                + fixed_point_matmul(w_hf, h_prev)
                + b_if
                + b_hf
            )
            i_t = sigmoid(
                fixed_point_matmul(w_ii, x)
                + fixed_point_matmul(w_hi, h_prev)
                + b_ii
                + b_hi
            )
            g_t = tanh(
                fixed_point_matmul(w_ig, x)
                + fixed_point_matmul(w_hg, h_prev)
                + b_ig
                + b_hg
            )
            o_t = sigmoid(
                fixed_point_matmul(w_io, x)
                + fixed_point_matmul(w_ho, h_prev)
                + b_io
                + b_ho
            )

            c_t = f_t * c_prev / fixed_point_one + i_t * g_t / fixed_point_one
            h_t = o_t * tanh(c_t) / fixed_point_one

            return h_t, c_t

        def q_inference(
            weights_and_bias: list[np.ndarray],
            inputs: np.ndarray,
            hidden_state: np.ndarray,
            cell_state: np.ndarray,
        ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
            inputs = np.array(inputs)

            for i in range(len(inputs[0])):
                hidden_state, cell_state = qlstm_cell(
                    inputs[0][i], hidden_state, cell_state, weights_and_bias
                )

            linear_w = to_fixed_point(np.expand_dims(weights_and_bias[4], axis=0))
            linear_b = to_fixed_point(np.expand_dims(weights_and_bias[5], axis=0))

            output = np.matmul(linear_w, hidden_state) / fixed_point_one + linear_b

            return output, hidden_state, cell_state

        try:
            test_mse = []
            for test_inputs, test_labels in test_loader:
                hidden_state = np.zeros([20, 1])
                cell_state = np.zeros([20, 1])

                test_outputs, hidden_state, cell_state = q_inference(
                    weights_and_bias, test_inputs, hidden_state, cell_state
                )

                mse = (to_float(test_outputs) - np.array(test_labels)) ** 2
                test_mse.append(mse)

            avg_mse = np.mean(test_mse)
            frac_bits_mse.append(avg_mse)
        except ValueError:
            break

    min_mse_idx = np.argmin(frac_bits_mse)
    best_frac_bits = frac_bits_choices[min_mse_idx]

    return FixedPoint.get_factory(total_bits=total_bits, frac_bits=best_frac_bits)


def main() -> None:
    args = read_commandline_args()

    model = LSTMModel()

    work_library_name = "xil_defaultlib"
    fixed_point_factory = determine_optimal_fixed_point_factory(
        model=model,
        saved_model_path="elasticai/creator/examples/assets/determine_fixed_point/model_v66",
        data_path="elasticai/creator/examples/assets/determine_fixed_point/pems-4w.csv",
        sensor_idx=4291,
        batch_size=1,
        datapoints_per_sample=6,
    )
    translation_args = dict(
        LSTM=LSTMTranslationArgs(
            fixed_point_factory=fixed_point_factory,
            sigmoid_resolution=(-2.5, 2.5, 256),
            tanh_resolution=(-1, 1, 256),
            work_library_name=work_library_name,
        ),
        Linear=Linear1dTranslationArgs(
            fixed_point_factory=fixed_point_factory,
            work_library_name=work_library_name,
        ),
    )

    code_repr = translator.translate_model(
        model=model,
        translation_args=translation_args,
        build_function_mapping=DEFAULT_BUILD_FUNCTION_MAPPING,
    )

    translator.save_code(code_repr=code_repr, path=args.build_dir)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Callable

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    fixed_point_params_from_factory,
)


@dataclass
class FPReLUComponent:
    fixed_point_factory: Callable[[float], FixedPoint]

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = fixed_point_params_from_factory(
            self.fixed_point_factory
        )

    @property
    def file_name(self) -> str:
        return "fp_relu.vhd"

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "fp_relu.tpl.vhd")

        code = template.format(
            data_width=self.data_width,
            clock_option="true",
        )

        yield from code.splitlines()



from typing import Any

import torch

from elasticai.creator.vhdl.number_representations import (
    FixedPointFactory,
    fixed_point_params_from_factory,
)


class FixedPointQuantFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> tuple[torch.Tensor, ...]:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(x: torch.Tensor, fixed_point_factory: Callable[[float], FixedPoint])"
            )
        x: torch.Tensor = args[0]
        fixed_point_factory: FixedPointFactory = args[1]

        total_bits, frac_bits = fixed_point_params_from_factory(fixed_point_factory)
        largest_fp_int = 2 ** (total_bits - 1) - 1
        fp_value = (x * (1 << frac_bits)).int().float()
        return fp_value.clamp(-largest_fp_int, largest_fp_int)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None


class FixedPointDequantFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> tuple[torch.Tensor, ...]:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(x: torch.Tensor, fixed_point_factory: Callable[[float], FixedPoint])"
            )
        x: torch.Tensor = args[0]
        fixed_point_factory: FixedPointFactory = args[1]

        total_bits, frac_bits = fixed_point_params_from_factory(fixed_point_factory)
        min_value = 2 ** (total_bits - frac_bits - 1) * (-1)
        max_value = (2 ** (total_bits - 1) - 1) / (1 << frac_bits)
        return (x / (1 << frac_bits)).clamp(min_value, max_value)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None



import unittest
from typing import Callable

import torch

from elasticai.creator.vhdl.evaluators.metric_evaluator import MetricEvaluator


class InferenceEvaluatorMock:
    def __init__(self, predictions: torch.Tensor, labels: torch.Tensor) -> None:
        self.predictions = predictions
        self.labels = labels

    def run(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.predictions, self.labels


def get_metric_evaluator(
    predictions: torch.Tensor,
    labels: torch.Tensor,
    metric: Callable[[torch.Tensor, torch.Tensor], torch.Tensor | float],
) -> MetricEvaluator:
    return MetricEvaluator(
        inference_evaluator=InferenceEvaluatorMock(predictions, labels), metric=metric
    )


def difference(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    return a - b


def summed_difference(a: torch.Tensor, b: torch.Tensor) -> float:
    return (a - b).sum().item()


class MetricEvaluatorTest(unittest.TestCase):
    def test_metric_evaluator_summed_differences_empty_predictions_labels(self) -> None:
        evaluator = get_metric_evaluator(
            predictions=torch.tensor([]),
            labels=torch.tensor([]),
            metric=summed_difference,
        )
        actual = evaluator.run()
        target = 0
        self.assertEqual(target, actual)

    def test_metric_evaluator_differences(self) -> None:
        evaluator = get_metric_evaluator(
            predictions=torch.tensor([1, 4, 3]),
            labels=torch.tensor([2, 2, 3]),
            metric=difference,
        )
        actual = evaluator.run().detach().numpy().tolist()
        target = [-1, 2, 0]
        self.assertEqual(target, actual)

    def test_metric_evaluator_summed_differences(self) -> None:
        evaluator = get_metric_evaluator(
            predictions=torch.tensor([1, 4, 3]),
            labels=torch.tensor([2, 2, 3]),
            metric=summed_difference,
        )
        actual = evaluator.run()
        target = 1
        self.assertEqual(target, actual)



import sys

import torch

from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory
from elasticai.creator.vhdl.quantized_modules.hard_sigmoid import FixedPointHardSigmoid
from elasticai.creator.vhdl.quantized_modules.linear import FixedPointLinear
from elasticai.creator.vhdl.quantized_modules.relu import FixedPointReLU
from elasticai.creator.vhdl.translator.abstract.layers.fp_hard_sigmoid_module import (
    FPHardSigmoidTranslationArgs,
)
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUTranslationArgs,
)
from elasticai.creator.vhdl.translator.abstract.layers.linear_1d_module import (
    Linear1dTranslationArgs,
)
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)
from elasticai.creator.vhdl.translator.pytorch.build_functions.linear_1d_build_function import (
    build_linear_1d,
)


class FixedPointModel(torch.nn.Module):
    def __init__(self, fixed_point_factory: FixedPointFactory) -> None:
        super().__init__()

        # self.linear1 = FixedPointLinear(
        #     in_features=2, out_features=3, fixed_point_factory=fixed_point_factory
        # )
        self.linear2 = FixedPointLinear(
            in_features=3, out_features=1, fixed_point_factory=fixed_point_factory
        )
        self.hard_sigmoid = FixedPointHardSigmoid(
            fixed_point_factory=fixed_point_factory
        )
        self.relu1 = FixedPointReLU(fixed_point_factory=fixed_point_factory)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.hard_sigmoid(self.relu1(self.linear2(x)))


def get_custom_build_mapping() -> BuildFunctionMapping:
    return DEFAULT_BUILD_FUNCTION_MAPPING.join_with_dict(
        {
            "elasticai.creator.vhdl.quantized_modules.linear.FixedPointLinear": build_linear_1d,
        }
    )


def main() -> None:
    if len(sys.argv) < 2:
        print("Please supply a build directory path as a program argument.")
        return
    build_path = sys.argv[1]

    fixed_point_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)

    model = FixedPointModel(fixed_point_factory)

    translation_args = dict(
        FixedPointLinear=Linear1dTranslationArgs(
            fixed_point_factory=fixed_point_factory, work_library_name="work"
        ),
        FixedPointHardSigmoid=FPHardSigmoidTranslationArgs(
            fixed_point_factory=fixed_point_factory
        ),
        FixedPointReLU=FPReLUTranslationArgs(fixed_point_factory=fixed_point_factory),
    )

    code_repr = translator.translate_model(
        model=model,
        translation_args=translation_args,
        build_function_mapping=get_custom_build_mapping(),
    )

    translator.save_code(code_repr=code_repr, path=build_path)


if __name__ == "__main__":
    main()



from typing import Any

import torch
import torch.fx as fx

from elasticai.creator.qat.layers import Binarize


class MyModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.test_lstm = torch.nn.LSTM(input_size=1, hidden_size=10)
        self.batch_norm = torch.nn.BatchNorm1d(2)

    def forward(self, x):
        return self.test_lstm(x)


class LSTMModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.lstm_container = MyModule()
        self.linear = torch.nn.Linear(in_features=10, out_features=1)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        return self.linear(self.lstm_container(x)[0])


if __name__ == "__main__":
    model = LSTMModel()

    tracer = fx.Tracer()
    graph = tracer.trace(model)

    new_graph = fx.Graph()
    node_args: dict[str, fx.Node] = {}

    def transform_argument_for_old_node_to_new_node(node_argument: fx.Node):
        return node_args[node_argument.name]

    gm: fx.GraphModule = fx.GraphModule(model, graph)

    def prepare_arguments_for_new_node(arg: Any) -> Any:
        if isinstance(arg, fx.Node):
            return node_args[arg.name]
        return arg

    for node in graph.nodes:
        if node.op == "output":
            new_graph.output(node_args["output"])
        else:
            if node.op == "call_module" and node.target == "linear":
                model.binarize = Binarize()
                new_node = new_graph.call_module(
                    module_name="binarize",
                    args=tuple(
                        [prepare_arguments_for_new_node(arg) for arg in node.args]
                    ),
                )
                node_args[node.name] = new_node
                new_node = new_graph.call_module(
                    "lstm_container.batch_norm", args=(new_node,)
                )
                node_args["output"] = new_node
            else:
                new_node = new_graph.node_copy(
                    node, transform_argument_for_old_node_to_new_node
                )
                node_args[node.name] = new_node
    print(new_graph)

    gm = fx.GraphModule(model, new_graph)

    print(gm)
    print(gm.graph)
    gm.graph.lint()

    print(tracer.trace(gm))
    gm(torch.tensor([[[1.0], [1.0]]]))



from typing import Callable

import torch
from torch.utils.data import Dataset

from elasticai.creator.vhdl.evaluators.evaluator import Evaluator
from elasticai.creator.vhdl.quantized_modules.typing import QuantType


def _run_inference_on_dataset(
    data: Dataset, predict_sample_fn: Callable[[torch.Tensor], torch.Tensor]
) -> tuple[torch.Tensor, torch.Tensor]:
    def to_tensor(x: list[torch.Tensor]) -> torch.Tensor:
        return torch.stack(x) if len(x) > 0 else torch.tensor([])

    predictions, labels = [], []
    for sample, label in data:
        prediction = predict_sample_fn(sample)
        predictions.append(prediction)
        labels.append(label)
    return to_tensor(predictions), to_tensor(labels)


class FullPrecisisonInferenceEvaluator(Evaluator):
    def __init__(self, module: torch.nn.Module, data: Dataset) -> None:
        self.module = module
        self.data = data

    def run(self) -> tuple[torch.Tensor, torch.Tensor]:
        return _run_inference_on_dataset(self.data, self.module)


class QuantizedInferenceEvaluator(Evaluator):
    def __init__(
        self,
        module: torch.nn.Module,
        data: Dataset,
        input_quant: QuantType = lambda x: x,
        output_dequant: QuantType = lambda x: x,
    ) -> None:
        if not hasattr(module, "quantized_forward"):
            raise ValueError(
                "`module` must be an object that has a `qunatized_forward` function."
            )
        self.module = module
        self.data = data
        self.input_quant = input_quant
        self.output_dequant = output_dequant

    def run(self) -> tuple[torch.Tensor, torch.Tensor]:
        def predict_sample(sample: torch.Tensor) -> torch.Tensor:
            quantized_data = self.input_quant(sample)
            prediction = self.module.quantized_forward(quantized_data)  # type: ignore
            return self.output_dequant(prediction)

        return _run_inference_on_dataset(self.data, predict_sample)



import unittest

from elasticai.creator.vhdl.components.fp_relu_component import FPReLUComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUModule,
    FPReLUTranslationArgs,
)


class FPReluModuleTest(unittest.TestCase):
    def setUp(self) -> None:
        self.module = FPReLUModule()
        self.fp_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
        self.args = FPReLUTranslationArgs(fixed_point_factory=self.fp_factory)

    def test_components_return_only_one_component(self) -> None:
        components = list(self.module.components(self.args))
        self.assertEqual(len(components), 1)
        self.assertEqual(type(components[0]), FPReLUComponent)

    def test_components_component_args_are_correctly_set(self) -> None:
        component = list(self.module.components(self.args))[0]
        self.assertEqual(component.fixed_point_factory, self.fp_factory)  # type: ignore



from elasticai.creator.vhdl.quantized_modules.hard_sigmoid import FixedPointHardSigmoid
from elasticai.creator.vhdl.translator.abstract.layers.fp_hard_sigmoid_module import (
    FPHardSigmoidModule,
)


def build_fp_hard_sigmoid(hard_sigmoid: FixedPointHardSigmoid) -> FPHardSigmoidModule:
    return FPHardSigmoidModule()



from collections.abc import Mapping
from typing import Any, Callable, Iterator

from elasticai.creator.vhdl.vhdl_component import VHDLModule

BuildFunction = Callable[[Any], VHDLModule]


class BuildFunctionMapping(Mapping[str, BuildFunction]):
    def __init__(self, mapping: dict[str, BuildFunction]):
        self._mapping = mapping

    def __getitem__(self, key: str) -> BuildFunction:
        return self._mapping[key]

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapping)

    @staticmethod
    def _get_cls_name(obj: Any) -> str:
        return f"{type(obj).__module__}.{type(obj).__name__}"

    def get_from_layer(self, layer: Any) -> BuildFunction | None:
        return self._mapping.get(self._get_cls_name(layer))

    def to_dict(self) -> dict[str, BuildFunction]:
        return self._mapping.copy()

    def join_with_dict(self, other: dict[str, BuildFunction]) -> "BuildFunctionMapping":
        """
        Join this build function mapping with another mapping and returns the new resulting mapping. If the current
        mapping and the other mapping share keys, the other mapping keys will overwrite the current keys.

        Parameters:
            other (dict[str, BuildFunction]): Dictionary that maps the path of a module to its BuildFunction.

        Returns:
            BuildFunctionMapping: The resulting build function mapping after joining both mappings.

        """
        return BuildFunctionMapping(mapping=self._mapping | other)



import torch
import torch.nn.functional as F

from elasticai.creator.vhdl.number_representations import (
    FixedPointFactory,
    fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.quantized_modules.autograd_functions import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.vhdl.quantized_modules.typing import QuantType


class _HardSigmoidBase(torch.nn.Hardsigmoid):
    def __init__(
        self,
        input_quant: QuantType = lambda x: x,
        input_dequant: QuantType = lambda x: x,
        inplace: bool = False,
    ) -> None:
        super().__init__(inplace)
        self.input_quant = input_quant
        self.input_dequant = input_dequant

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        q_x = self.input_dequant(self.input_quant(x))
        return F.hardsigmoid(q_x)

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("The quantized_forward function is not implemented.")


class FixedPointHardSigmoid(_HardSigmoidBase):
    def __init__(
        self, fixed_point_factory: FixedPointFactory, inplace: bool = False
    ) -> None:
        super().__init__(
            input_quant=lambda x: FixedPointQuantFunction.apply(x, fixed_point_factory),
            input_dequant=lambda x: FixedPointDequantFunction.apply(
                x, fixed_point_factory
            ),
            inplace=inplace,
        )
        self.fixed_point_factory = fixed_point_factory

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        total_bits, frac_bits = fixed_point_params_from_factory(
            self.fixed_point_factory
        )

        def fp(value: float) -> int:
            largest_fp_int = 2 ** (total_bits - 1) - 1
            int_value = int(value * (1 << frac_bits))
            return min(max(int_value, -largest_fp_int), largest_fp_int)

        def fp_hard_sigmoid(a: int) -> int:
            if a <= fp(-3.0):
                return 0
            elif a >= fp(3.0):
                return fp(1.0)
            else:
                return int(a * fp(1.0 / 6.0) / fp(1.0)) + fp(0.5)

        return x.detach().apply_(fp_hard_sigmoid)



from dataclasses import dataclass, field
from typing import Callable

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.components.utils import (
    calculate_addr_width,
    derive_fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint


@dataclass
class Linear1dComponent:
    in_features: int
    out_features: int
    fixed_point_factory: Callable[[float], FixedPoint]
    work_library_name: str = field(default="work")

    def __post_init__(self) -> None:
        if self.out_features != 1:
            raise NotImplementedError(
                "Currently only one bias is supported (which implies that out_features must be 1)."
            )

        self.data_width, _ = derive_fixed_point_params_from_factory(
            self.fixed_point_factory
        )
        self.addr_width = calculate_addr_width(self.in_features)

    @property
    def file_name(self) -> str:
        return "linear_1d.vhd"

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "linear_1d.tpl.vhd")

        code = template.format(
            work_library_name=self.work_library_name,
            addr_width=self.addr_width,
            data_width=self.out_features,
            in_feature_count=self.in_features,
            out_feature_count=self.out_features,
        )

        yield from code.splitlines()



from typing import Callable

import torch

from elasticai.creator.vhdl.number_representations import (
    FixedPointFactory,
    fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.quantized_modules.autograd_functions import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.vhdl.quantized_modules.typing import QuantType

OperationType = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


class _LinearBase(torch.nn.Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        matmul_op: OperationType = torch.matmul,
        add_op: OperationType = torch.add,
        input_quant: QuantType = lambda x: x,
        input_dequant: QuantType = lambda x: x,
        param_quant: QuantType = lambda x: x,
        param_dequant: QuantType = lambda x: x,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__(in_features, out_features, bias, device, dtype)
        self.matmul_op = matmul_op
        self.add_op = add_op
        self.input_quant = input_quant
        self.input_dequant = input_dequant
        self.param_quant = param_quant
        self.param_dequant = param_dequant

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        dq_input = self.input_dequant(self.input_quant(x))
        dq_weight = self.param_dequant(self.param_quant(self.weight))

        if self.bias is not None:
            dq_bias = self.param_dequant(self.param_quant(self.bias))
            return self.add_op(self.matmul_op(dq_input, dq_weight.T), dq_bias)

        return self.matmul_op(dq_input, dq_weight.T)

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("The quantized_forward function is not implemented.")


class FixedPointLinear(_LinearBase):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        fixed_point_factory: FixedPointFactory,
        bias: bool = True,
        device=None,
    ) -> None:
        self.fixed_point_factory = fixed_point_factory
        self.quant = lambda x: FixedPointQuantFunction.apply(x, fixed_point_factory)
        self.dequant = lambda x: FixedPointDequantFunction.apply(x, fixed_point_factory)
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            input_quant=self.quant,
            input_dequant=self.dequant,
            param_quant=self.quant,
            param_dequant=self.dequant,
            device=device,
        )

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        total_bits, frac_bits = fixed_point_params_from_factory(
            self.fixed_point_factory
        )

        def fp_matmul(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
            return (torch.matmul(a, b) / (1 << frac_bits)).int().float()

        def clamp_overflowing_values(a: torch.Tensor) -> torch.Tensor:
            largest_fp_int = 2 ** (total_bits - 1) - 1
            return a.clamp(-largest_fp_int, largest_fp_int)

        q_weight = self.param_quant(self.weight)
        output = fp_matmul(x, q_weight.T)

        if self.bias is not None:
            q_bias = self.param_quant(self.bias)
            output = self.add_op(output, q_bias)

        return clamp_overflowing_values(output)



"""Interfaces/Adapters to the used machine learning framework"""

from typing import Any, Callable, Iterable, Iterator, Protocol, runtime_checkable

from numpy.typing import ArrayLike

Index = int | slice | tuple[int | slice]


class Tensor(Protocol):
    def __getitem__(self, index: Index) -> "Tensor":
        ...

    def apply_(self, callable: Callable) -> "Tensor":
        ...

    def detach(self) -> "Tensor":
        ...

    def numpy(self) -> ArrayLike:
        ...


@runtime_checkable
class Parameter(Tensor, Protocol):
    ...


@runtime_checkable
class Module(Protocol):
    @property
    def training(self) -> bool:
        ...

    def extra_repr(self) -> str:
        ...

    def named_children(self) -> Iterable[tuple[str, "Module"]]:
        ...

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        ...

    def state_dict(self) -> dict[str, Any]:
        ...

    def __call__(self, x: Any, *args: Any, **kwargs: Any) -> Any:
        ...



from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch.build_functions import (
    build_fp_hard_sigmoid,
    build_fp_relu,
    build_linear_1d,
    build_lstm,
)

DEFAULT_BUILD_FUNCTION_MAPPING = BuildFunctionMapping(
    mapping={
        "torch.nn.modules.rnn.LSTM": build_lstm,
        "torch.nn.modules.linear.Linear": build_linear_1d,
        "elasticai.creator.vhdl.quantized_modules.hard_sigmoid.FixedPointHardSigmoid": build_fp_hard_sigmoid,
        "elasticai.creator.vhdl.quantized_modules.relu.FixedPointReLU": build_fp_relu,
    }
)



import unittest
from dataclasses import dataclass
from typing import Any, Iterable

import torch

from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import LSTMModule
from elasticai.creator.vhdl.translator.abstract.layers.linear_1d_module import (
    Linear1dTranslationArgs,
)
from elasticai.creator.vhdl.translator.abstract.layers.lstm_module import (
    LSTMTranslationArgs,
)
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)
from elasticai.creator.vhdl.translator.pytorch.translator import CodeFile, CodeModule
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


@dataclass
class VHDLComponentMock(VHDLComponent):
    name: str
    code: list[str]

    @property
    def file_name(self) -> str:
        return self.name

    def __call__(self) -> Code:
        yield from self.code


@dataclass
class VHDLModuleMock(VHDLModule):
    vhdl_components: list[VHDLComponent]

    def components(self, args: Any) -> Iterable[VHDLComponent]:
        yield from self.vhdl_components


def fake_build_function(module: torch.nn.Module) -> VHDLModuleMock:
    return VHDLModuleMock(
        vhdl_components=[
            VHDLComponentMock(name="component1", code=["1", "2", "3"]),
            VHDLComponentMock(name="component2", code=["4", "5", "6"]),
        ]
    )


def unpack_module_directories(
    modules: Iterable[CodeModule],
) -> list[tuple[str, list[tuple[str, Code]]]]:
    def unpack_code_file(code_file: CodeFile) -> tuple[str, Code]:
        return code_file.file_name, list(code_file.code)

    return [
        (module.module_name, list(map(unpack_code_file, module.files)))
        for module in modules
    ]


class TranslatorTest(unittest.TestCase):
    def setUp(self) -> None:
        self.build_mapping = BuildFunctionMapping(
            mapping={"torch.nn.modules.rnn.LSTM": fake_build_function}
        )

    def test_translate_model_empty_model(self) -> None:
        model = torch.nn.Sequential()
        generated_code = translator.translate_model(
            model, translation_args=dict(), build_function_mapping=self.build_mapping
        )
        self.assertEqual(len(list(generated_code)), 0)

    def test_generate_code(self) -> None:
        model = torch.nn.Sequential(torch.nn.LSTM(input_size=1, hidden_size=2))
        code_containers = translator.translate_model(
            model, translation_args=dict(), build_function_mapping=self.build_mapping
        )

        code = unpack_module_directories(code_containers)
        expected_code = [
            (
                "0_LSTM",
                [
                    ("component1", ["1", "2", "3"]),
                    ("component2", ["4", "5", "6"]),
                ],
            )
        ]

        self.assertEqual(code, expected_code)

    @unittest.SkipTest
    def test_translate_model_correct_ordering_of_layers(self):
        class Model(torch.nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.lstm = torch.nn.LSTM(input_size=1, hidden_size=2)
                self.linear = torch.nn.Linear(in_features=2, out_features=1)

            def forward(self, x: torch.Tensor) -> torch.Tensor:
                return self.lstm(self.linear(x))

        model = Model()
        translated = list(
            translator.translate_model(
                model,
                translation_args=dict(
                    Linear=Linear1dTranslationArgs(
                        fixed_point_factory=FixedPoint.get_factory(8, 4)
                    ),
                    LSTM=LSTMTranslationArgs(
                        fixed_point_factory=FixedPoint.get_factory(8, 4),
                        sigmoid_resolution=(-2.5, 2.5, 100),
                        tanh_resolution=(-2.5, 2.5, 100),
                    ),
                ),
                build_function_mapping=DEFAULT_BUILD_FUNCTION_MAPPING,
            )
        )

        self.assertEqual(translated[0].module_name, "0_Linear")
        self.assertEqual(translated[1].module_name, "1_LSTM")



from collections.abc import Callable
from copy import deepcopy
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset, TensorDataset

from elasticai.creator.vhdl.evaluators.inference_evaluator import (
    QuantizedInferenceEvaluator,
)
from elasticai.creator.vhdl.evaluators.metric_evaluator import MetricEvaluator
from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory
from elasticai.creator.vhdl.quantized_modules import (
    FixedPointHardSigmoid,
    FixedPointLinear,
    FixedPointReLU,
)
from elasticai.creator.vhdl.quantized_modules.autograd_functions import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)


def get_dataset() -> tuple[torch.Tensor, torch.Tensor]:
    """
    Retruns the inputs and outputs of the following logic function:
        O = (!I1 * I2 * I3) + (I1 * !I2 * I3) + (I1 * I2 * !I3)
    """
    x = torch.as_tensor(
        [
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
        ],
        dtype=torch.float32,
    )
    y = torch.as_tensor(
        [
            [0],
            [0],
            [0],
            [1],
            [0],
            [1],
            [1],
            [0],
        ],
        dtype=torch.float32,
    )
    return x, y


def augment_data(
    samples: torch.Tensor,
    labels: torch.Tensor,
    expand_factor: int,
    noise_var: float,
    seed: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    augmented_samples = samples.tile(expand_factor, 1)
    augmented_labels = labels.tile(expand_factor, 1)
    torch.random.manual_seed(seed)
    noise = torch.randn_like(augmented_samples) * noise_var
    augmented_samples = augmented_samples + noise
    return augmented_samples, augmented_labels


def binary_accuracy(
    predictions: torch.Tensor, labels: torch.Tensor, threshold: float
) -> float:
    binary_predictions = torch.where(predictions >= threshold, 1.0, 0.0)
    true_positives = sum(
        bool(pred == label) for pred, label in zip(binary_predictions, labels)
    )
    return true_positives / len(predictions)


def train(
    model: torch.nn.Module,
    ds_train: Dataset,
    ds_val: Dataset,
    batch_size: int,
    learning_rate: float,
    num_epochs: int,
    fp_factory: FixedPointFactory,
) -> tuple[list[float], list[float], list[float]]:
    dl_train = DataLoader(ds_train, batch_size=batch_size, shuffle=True)
    dl_val = DataLoader(ds_val, batch_size=batch_size)

    loss_fn = torch.nn.BCELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    accuracy_metric: Callable[[torch.Tensor], torch.Tensor] = partial(
        binary_accuracy, threshold=0.5
    )

    quantized_inference_evaluator = QuantizedInferenceEvaluator(
        module=model,
        data=ds_val,
        input_quant=lambda x: FixedPointQuantFunction.apply(x, fp_factory),
        output_dequant=lambda x: FixedPointDequantFunction.apply(x, fp_factory),
    )
    quantized_loss_evaluator = MetricEvaluator(
        inference_evaluator=quantized_inference_evaluator, metric=loss_fn
    )
    quantized_accuracy_evaluator = MetricEvaluator(
        inference_evaluator=quantized_inference_evaluator, metric=accuracy_metric
    )

    train_loss_per_epoch = []
    val_loss_per_epoch = []
    val_accuracy_per_epoch = []

    for epoch in range(num_epochs):
        model.train(True)

        running_train_loss = 0.0

        for inputs, labels in dl_train:
            optimizer.zero_grad()
            outputs = model(inputs)

            loss = loss_fn(outputs, labels)
            loss.backward()

            optimizer.step()
            running_train_loss += loss.item()

        train_loss_per_epoch.append(running_train_loss / len(dl_train))

        model.train(False)

        running_val_loss = 0.0
        running_accuracy = 0.0

        for inputs, labels in dl_val:
            outputs = model(inputs)
            loss = loss_fn(outputs, labels)

            running_val_loss += loss.item()
            running_accuracy += accuracy_metric(outputs, labels)

        val_loss_per_epoch.append(running_val_loss / len(dl_val))
        val_accuracy_per_epoch.append(running_accuracy / len(dl_val))

        print(
            f"[epoch {epoch + 1}]",
            f"loss: {train_loss_per_epoch[epoch]:.04};",
            f"val_loss: {val_loss_per_epoch[epoch]:.04};",
            f"val_acc: {val_accuracy_per_epoch[epoch]:.04}",
        )

    simulated_quant_val_loss = quantized_loss_evaluator.run()
    simulated_quant_val_acc = quantized_accuracy_evaluator.run()

    print(
        "[training summary]",
        f"\tloss: {train_loss_per_epoch[-1]:.04};",
        f"\tval_loss: {val_loss_per_epoch[-1]:.04};",
        f"\tval_acc: {val_accuracy_per_epoch[-1]:.04}",
        f"\tsimulated_quant_val_loss: {simulated_quant_val_loss:.04};",
        f"\tsimulated_quant_val_acc: {simulated_quant_val_acc:.04}",
        sep="\n",
    )

    return train_loss_per_epoch, val_loss_per_epoch, val_accuracy_per_epoch


def plot_params(init_model: torch.nn.Module, final_model: torch.nn.Module) -> None:
    def get_params(model: torch.nn.Module) -> np.ndarray:
        return np.concatenate(
            [param.detach().numpy().flatten() for param in model.parameters()]
        )

    init_params = get_params(init_model)
    final_params = get_params(final_model)
    param_indices = np.arange(len(init_params))

    bar_width = 0.4
    plt.bar(param_indices - 0.2, init_params, bar_width, label="initial params")
    plt.bar(param_indices + 0.2, final_params, bar_width, label="final params")
    plt.legend()
    plt.show()


def plot_loss_curve(
    train_losses: list[float], val_losses: list[float], val_accuracy: list[float]
) -> None:
    plt.plot(train_losses, label="train loss")
    plt.plot(val_losses, label="validation loss")
    plt.plot(val_accuracy, label="validation accuracy")
    plt.legend()
    plt.show()


class FixedPointModel(torch.nn.Module):
    def __init__(self, fixed_point_factory: FixedPointFactory) -> None:
        super().__init__()
        self._linear1 = FixedPointLinear(
            in_features=3, out_features=2, fixed_point_factory=fixed_point_factory
        )
        self._linear2 = FixedPointLinear(
            in_features=2, out_features=1, fixed_point_factory=fixed_point_factory
        )
        self._relu = FixedPointReLU(fixed_point_factory=fixed_point_factory)
        self._sigmoid = FixedPointHardSigmoid(fixed_point_factory=fixed_point_factory)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._linear1(x)
        x = self._relu(x)
        x = self._linear2(x)
        x = self._sigmoid(x)
        return x

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._linear1.quantized_forward(x)
        x = self._relu.quantized_forward(x)
        x = self._linear2.quantized_forward(x)
        x = self._sigmoid.quantized_forward(x)
        return x


def main() -> None:
    x, y = get_dataset()
    x_train, y_train = augment_data(x, y, expand_factor=240, noise_var=0.05, seed=24)
    x_test, y_test = augment_data(x, y, expand_factor=160, noise_var=0.05, seed=24)
    ds_train = TensorDataset(x_train, y_train)
    ds_test = TensorDataset(x_test, y_test)

    fixed_point_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
    final_model = FixedPointModel(fixed_point_factory)
    init_model = deepcopy(final_model)

    history = train(
        model=final_model,
        ds_train=ds_train,
        ds_val=ds_test,
        batch_size=16,
        learning_rate=1e-3,
        num_epochs=200,
        fp_factory=fixed_point_factory,
    )

    plot_loss_curve(*history)
    plot_params(init_model, final_model)


if __name__ == "__main__":
    main()



import unittest

from elasticai.creator.vhdl.components.fp_relu_component import FPReLUComponent
from elasticai.creator.vhdl.number_representations import FixedPoint


class FPReLUComponentTest(unittest.TestCase):
    def test_relu_correct_number_of_lines(self) -> None:
        to_fp = FixedPoint.get_factory(total_bits=8, frac_bits=4)

        component = FPReLUComponent(
            fixed_point_factory=to_fp,
        )

        self.assertEqual(len(list(component())), 66)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import numpy as np
from torch.utils.data import Dataset

Slice = tuple[int | None, int | None]


@dataclass
class PeMSDataset(Dataset):
    """
    The dataset can be downloaded from this site (pems-4w.csv is used): https://zenodo.org/record/3939793
    """

    dataset_path: str | Path
    datapoints_per_sample: int
    sensor_idx: int
    raw_data_slice: Optional[Slice] = field(default=None)

    def __post_init__(self) -> None:
        assert (
            self.datapoints_per_sample > 0
        ), "datapoints_per_sample must be greater than 0."
        self.data = self._load_data()

    def __len__(self) -> int:
        n_samples = len(self.data) - (self.datapoints_per_sample + 1) + 1
        return n_samples if n_samples > 0 else 0

    def __getitem__(self, idx: int) -> tuple[np.ndarray, np.ndarray]:
        sample = self.data[idx : idx + self.datapoints_per_sample]
        sample = sample.reshape((self.datapoints_per_sample, 1))
        label = self.data[idx + self.datapoints_per_sample]
        label = np.array([label], dtype=self.data.dtype)
        return sample, label

    def _load_data(self) -> np.ndarray:
        full_data = np.loadtxt(self.dataset_path, delimiter=",", dtype=np.float32)
        sensor_data = full_data[self.sensor_idx]
        if self.raw_data_slice is not None:
            start, end = self.raw_data_slice
            sensor_data = sensor_data[start:end]
        return sensor_data
